// Copyright 1996-2008 by Jon Dart.  All Rights Reserved.
#ifndef _BITBOARD_H
#define _BITBOARD_H

#include "types.h"
#include <iostream>
#ifdef USE_INTRINSICS
#include <intrin.h>
#endif

class Bitboard
{
       friend ostream & operator << (ostream &o, const Bitboard &b);

#define GETBIT64(x) (uint64)(((int64)x)&-((int64)x))
#define GETBIT32(x) (uint32)(((int32)x)&-((int32)x))

    public:	

       struct ints
       {
          uint32 loval,hival;
       };
       struct shorts
       {
          uint16 lo1,lo2,hi1,hi2;
       };
       union conv
       {
          uint64 val1;
          ints val2;
          shorts val3;
       };

    Bitboard()
    : data(0)
    {
    }
    
    Bitboard(uint64 n)
    : data(n)
    {
    }

    Bitboard(uint32 hi,uint32 lo) {
      ((conv*)&data)->val2.hival = hi;
      ((conv*)&data)->val2.loval = lo;
    }

    operator uint64() const
    {
       return data;
    }

    byte byte_value() const
    {
       // PC is little-endian
       return *((byte*)&data);
    }
    
    uint32 FORCEINLINE hivalue() const
    {
      return ((conv*)(this))->val2.hival;
    }
    
    uint32 FORCEINLINE lovalue() const
    {
      //return ((conv*)(this))->val2.loval;
      return (uint32)data;
    }
    
    void clear()
    {
       data = 0;
    }
    void set_lsb()
    {
       data &= ~(data-1);
    }
    void set(int n)
    {
#if defined(_WIN64) && defined(_MSC_VER) && defined(USE_INTRINSICS)
       _bittestandset64((LONG64*)&data,(DWORD)n);
#elif defined(_64BIT)
       data |= ((uint64)1)<<n;
#elif defined (_MSC_VER) && defined(USE_INTRINSICS)
       if (n<32)
         _bittestandset((LONG*)&data,n);
       else
         _bittestandset(((LONG*)&data)+1,n-32);
#else
       data |= mask[n];
#endif
    }
    void FORCEINLINE clear(int n)
    {
#if defined(_WIN64) && defined(_MSC_VER) && defined(USE_INTRINSICS)
       _bittestandreset64((LONG64*)&data,(DWORD)n);
#elif defined(_64BIT)
       data &= ~(((uint64)1)<<n);
#elif defined(_WIN32) && defined(USE_INTRINSICS)
       if (n<32) _bittestandreset((LONG*)&data,n);
       else _bittestandreset(((LONG*)&data)+1,n-32);
#else
       data &= ~mask[n];
#endif
    }
    void setClear(const Bitboard &b)
    {
        data ^= b.data;
    }
    int FORCEINLINE isSet(int n) const
    {
#if defined(_WIN64) & defined(USE_INTRINSICS)
        return _bittest64((__int64*)&data,n);
#elif defined(_WIN32) & defined(USE_INTRINSICS)
        if (n<32)
          return _bittest((const LONG*)&data,n);
        else
          return _bittest(((const LONG*)&data)+1,n-32);
#elif defined(_64BIT)
        return (data & (((uint64)1)<<n)) != (uint64)0;
#else
        return ((data & mask[n]) != (uint64)0);
#endif
    }
    int FORCEINLINE is_clear()const
    {
        return (data == (uint64)0);
    }
    Bitboard operator & (const Bitboard &src) const {
      return Bitboard(data & src.data);
    }
    Bitboard operator & (uint64 src) const {
      return Bitboard(data & src);
    }
    const Bitboard & operator &= (const Bitboard &src) {
      data &= src.data;
      return *this;
    }
    const Bitboard & operator &= (uint64 src) {
      data &= src;
      return *this;
    }
    Bitboard operator | (const Bitboard &src) const {
      return Bitboard(data | src.data);
    }
    Bitboard operator | (uint64 src) const {
      return Bitboard(data | src);
    }
    const Bitboard & operator |= (const Bitboard &src) {
      data |= src.data;
      return *this;
    }
    const Bitboard & operator |= (uint64 src) {
      data |= src;
      return *this;
    }

    void shr(int n) {
        data = data >> n;
    }

    static inline Bitboard shr(const Bitboard &src, int n) 
    {
        return Bitboard(src.data >> n);    
    }

    void shr8()
    {
#if defined(_64BIT)
        data >>= 8;
#else
        char *p = (char*)&data;
        byte b = p[4];
        ((conv*)&data)->val2.hival >>= 8;
        ((conv*)&data)->val2.loval >>= 8;
        p[3] = b;
#endif
    }

    static Bitboard shl(const Bitboard &src, int n)
    {
        return Bitboard(src.data << n);
    }

    void shl(int n)
    {
        data = data << n;    
    }

    void shl_byte(int n)
    {
        // shift lowest 8 bits left by n bytes
        char *p = (char*)&data;
        p[n] = *p;
        if (n>0) *p = 0;
    }
    static inline Bitboard shl_byte(byte b, int n)
    {
        // shift lowest 8 bits left by n bytes
        uint64 new_data = 0;
        char *p = (char*)&new_data;
        p[n] = b;
        return Bitboard(new_data);
    }

    void shl8()
    {
#if defined(_64BIT)
        data <<= 8;
#else
        char *p = (char*)&data;
        byte b = p[3];
        ((conv*)&data)->val2.loval <<= 8;
        ((conv*)&data)->val2.hival <<= 8;
        p[4] = b;
#endif
    }
    
    int operator == (const Bitboard &b) const
    {
       return b.data == data;
    }

    int operator != (const Bitboard &b) const
    {
       return b.data != data;
    }

    // extract the bits for the nth rank (from the Black side)
    unsigned rank_bits(int n) const
    {
       byte *p = (byte*)&data;
       return (unsigned)p[n];
    }

    // return the number of bits set
    FORCEINLINE unsigned int bitCount() const
    {
      /**
       unsigned register int n;
       uint64 tmp = data;
       for (n = 0; tmp != 0; n++, tmp &= (tmp - 1));
       return n;
      **/
       // see http://en.wikipedia.org/wiki/Hamming_weight
       uint64 x = data;
       x -= (x >> 1) & m1;
       x = (x & m2) + ((x >> 2) & m2);
       x = (x + (x >> 4)) & m4; 
       return (int)((x * h01)>>56); 
    }
    
    FORCEINLINE Square firstOne() const {
#if defined(_WIN64) && defined(_MSC_VER) && defined(USE_INTRINSICS)
      DWORD index;
      if (_BitScanForward64(&index,data))
        return (Square)index;
      else 
        return InvalidSquare;
#elif defined(_64BIT)
      if (data == 0) return InvalidSquare;
      else return MagicTable64[(GETBIT64(data)*MAGIC64)>>58];
#else
      if (data == 0) return InvalidSquare;
      // use De Bruijn multiplication code from Lasse Hansen
      if (lovalue())
        return MagicTable32[(GETBIT32(lovalue())*MAGIC32)>>27];
      else
        return MagicTable32[(GETBIT32(hivalue())*MAGIC32)>>27]+32;
#endif
    }
    
    FORCEINLINE int iterate(Square &sq) {
#if defined(_WIN64) && defined(_MSC_VER) && defined(USE_INTRINSICS)
      if (!_BitScanForward64((DWORD*)&sq,data)) {
         return 0;
      }
      else {
         _bittestandreset64((LONG64*)&data,(DWORD)sq);
         return 1;
      }
#elif defined(_64BIT)
      if (data == 0) return 0;
      uint64 tmp = (uint64)GETBIT64(data);
      sq = MagicTable64[(tmp*MAGIC64)>>58];
      // clear bit we will return
      data &= ~tmp;
#else
      if (data == 0) return 0;
#if defined(USE_INTRINSICS)
      unsigned long idx;
      unsigned int* pbb = (unsigned int*)&data;
      if (_BitScanForward(&idx, *pbb )) {
         *pbb = *pbb & (*pbb - 1);
         sq =(Square)idx;
		 return 1;
      }
      pbb++;
      _BitScanForward(&idx, *pbb);
      *pbb = *pbb & (*pbb - 1);
      sq =(Square)idx + 32;
#else
      // use De Bruijn multiplication code from Lasse Hansen
      uint32 tmp;
      if (lovalue()) {
        tmp = GETBIT32(lovalue());
        sq = MagicTable32[(tmp*MAGIC32)>>27];
        ((conv*)&data)->val2.loval &= ~tmp;
      }
      else {
        tmp = GETBIT32(hivalue());
        sq = MagicTable32[(tmp*MAGIC32)>>27]+32;
        ((conv*)&data)->val2.hival &= ~tmp;
      }
#endif
#endif
      return 1;
    }
    
    static void init();
    
    static void cleanup();

    static CACHE_ALIGN int MagicTable32[32];
#if defined(_64BIT)
    static CACHE_ALIGN int MagicTable64[64];
#endif
    static CACHE_ALIGN const uint64 mask[64];
    static CACHE_ALIGN int unrot90[64];
    static CACHE_ALIGN int R90[64];
    static CACHE_ALIGN int da1[64];
    static CACHE_ALIGN int da1mod8[64];
    static CACHE_ALIGN int da1div8[64];
    static CACHE_ALIGN int unrot_da1[64];
    static CACHE_ALIGN int da1_mask[64];
    static CACHE_ALIGN int da8[64];
    static CACHE_ALIGN int da8mod8[64];
    static CACHE_ALIGN int da8div8[64];
    static CACHE_ALIGN int unrot_da8[64];
    static CACHE_ALIGN int da8_mask[64];
    
    uint64 data;

    private:
#ifdef _64BIT
    static const uint64 MAGIC64 = 0x07EDD5E59A4E28C2;
#endif
    static const unsigned MAGIC32 = 0xe89b2be;
    static const uint64 m1  = 0x5555555555555555ULL; //binary: 0101...
    static const uint64 m2  = 0x3333333333333333ULL; //binary: 00110011..
    static const uint64 m4  = 0x0f0f0f0f0f0f0f0fULL; //binary:  4 zeros,  4 ones ...
    static const uint64 h01 = 0x0101010101010101ULL; //the sum of 256 to the power of 0,1,2,3...


};

#define DIAG_UPPER_LEFT 0
#define DIAG_UPPER_RIGHT 1
#define DIAG_LOWER_LEFT 2
#define DIAG_LOWER_RIGHT 3

struct DiagInfo2
{
   int shift;
   int byte_shift;
   int bit_shift;
   int bit;
   int mask;
   Bitboard *attacks;
   Bitboard upperMask;
   Bitboard lowerMask;
};

extern CACHE_ALIGN DiagInfo2 DiagInfoA1[64];
extern CACHE_ALIGN DiagInfo2 DiagInfoA8[64];

inline int TEST_MASK(const Bitboard &b1,const Bitboard &b2) {
  return (uint64)(b1 & b2) != (uint64)0;
}


#endif
